// contactmodelHill.cpp
#include "contactmodelHill.h"

#include "module/interface/ibody.h"
#include "module/interface/ibodymechanical.h"
#include "module/interface/icontactmechanical.h"
#include "module/interface/icontact.h"

#include "module/interface/ipiecemechanical.h"
#include "module/interface/ipiece.h"
#include "version.txt"

#include "module/interface/ifishcalllist.h"
#include "utility/src/tptr.h"
#include "base/src/mathutil.h"

#include "kernel/interface/iprogram.h"
#include "module/interface/icontactthermal.h"
#include "contactmodel/src/contactmodelthermal.h"

#ifdef Hill_LIB
    int __stdcall DllMain(void *,unsigned, void *) {
      return 1;
    }

    extern "C" EXPORT_TAG const char *getName() {
      return "contactmodelmechanical3dHill";
    }

    extern "C" EXPORT_TAG unsigned getMajorVersion() {
      return MAJOR_VERSION;
    }

    extern "C" EXPORT_TAG unsigned getMinorVersion() {
      return MINOR_VERSION;
    }

    extern "C" EXPORT_TAG void *createInstance() {
      cmodelsxd::ContactModelHill *m = new cmodelsxd::ContactModelHill();
      return (void *)m;
    }
#endif // Hill_LIB

namespace cmodelsxd {
    using namespace itasca;
  
    int ContactModelHill::index_ = -1;
    UInt ContactModelHill::getMinorVersion() const { return MINOR_VERSION;}

    ContactModelHill::ContactModelHill() : mu_(0.0)
                                         , psi_(0.0)
                                         , alpha_(0.0)
                                         , knSq_(0.0)
                                         , ksSq_(0.0)
                                         , etanSq_(0.0)
                                         , etasSq_(0.0)
                                         , moisState_(0)
                                         , del0_(0.0)
                                         , Fs_( DVect3(0.0) )
                                         , Fm_(0.0)
                                         , Fsh0_( DVect2(0.0) )
                                         , R0_(0.0) {
    }

    ContactModelHill::~ContactModelHill() {
    }

    bool ContactModelHill::getPropertyReadOnly(uint i) const {
      switch (i) {
        case kwYoungMod:
        case kwPoisRatio:
        case kwNStiffCoef:
        case kwSStiffCoef:
        case kwNDampCoef:
        case kwSDampCoef:
        case kwMoisState:
        case kwOverlap:
        case kwSurfForce:
        case kwMoisForce:
          return true;
        case kwFricCoef:
        case kwSuction:
        case kwDampCon:
          return false;
        default:
          break;
      }
      return true;
    }

    bool ContactModelHill::getPropertyGlobal(uint i) const {
      switch (i) {
        case kwSurfForce:
          return false;
        default:
          break;
      }
      return true;
    }

    bool ContactModelHill::supportsInheritance(uint i) const {
      switch (i) {
        case kwYoungMod:
        case kwPoisRatio:
          return true;
        default:
          break;
      }
      return false;
    }

    bool ContactModelHill::endPropertyUpdated( const QString &name, const IContactMechanical *c ) {
      assert(c);
      QStringList availableProperties = getProperties().split(",",QString::SkipEmptyParts);
      QRegExp rx(name,Qt::CaseInsensitive);
      int idx = availableProperties.indexOf(rx)+1;
      if ( idx <= 0 ) return false;
      switch(idx) {
        case kwYoungMod:
        case kwPoisRatio: {
          updateCoefs(c);
          return true;
        }
      }
      return false;
    }

    QVariant ContactModelHill::getProperty(uint i,const IContact *con) const {
      QVariant var;
      switch (i) {
        case kwFricCoef:   return mu_;
        case kwSuction:    return psi_;
        case kwDampCon:    return alpha_;
        case kwYoungMod:   return -1.0;
        case kwPoisRatio:  return -1.0;
        case kwNStiffCoef: return knSq_;
        case kwSStiffCoef: return ksSq_;
        case kwNDampCoef:  return etanSq_;
        case kwSDampCoef:  return etasSq_;
        case kwMoisState:  return moisState_;
        case kwOverlap:    return del0_;
        case kwSurfForce:  var.setValue( Fs_ ); return var;
        case kwMoisForce:  return Fm_;
      }
      assert(0);
      return QVariant();
    }

    bool ContactModelHill::setProperty(uint i,const QVariant &v,IContact *) {
      switch (i) {
        case kwFricCoef: {
          if ( !v.canConvert<double>() )
            throw Exception("fric_coef must be a double.");
          double val( v.toDouble() );
          if ( val < 0.0 )
            throw Exception("Negative fric_coef not allowed.");
          mu_ = val;
          return false;
        }
        case kwSuction: {
          if ( !v.canConvert<double>() )
            throw Exception("suction must be a double.");
          double val( v.toDouble() );
          if ( val < 0.0 )
            throw Exception("Negative suction not allowed.");
          psi_ = val;
          return false;
        }
        case kwDampCon: {
          if ( !v.canConvert<double>() )
            throw Exception("damp_con must be a double.");
          double val( v.toDouble() );
          if ( (val < 0.0) || (val > 1.0) )
            throw Exception("damp_con must be in range [0.0, 1.0].");
          alpha_ = val;
          return true;
        }
      }
      return false;
    }

    void ContactModelHill::archive(ArchiveStream &stream) {
      stream & mu_;
      stream & psi_;
      stream & alpha_;
      stream & knSq_;
      stream & ksSq_;
      stream & etanSq_;
      stream & etasSq_;
      stream & moisState_;
      stream & del0_;
      stream & Fs_;
      stream & Fm_;
      stream & Fsh0_;
      stream & R0_;
    }

    void ContactModelHill::copy(const ContactModel *cm) {
      ContactModelMechanical::copy(cm);
      const ContactModelHill *in = dynamic_cast<const ContactModelHill*>(cm);
      if (!in) throw std::runtime_error("Internal error: contact model dynamic cast failed.");
      mu_        = in->mu_;
      psi_       = in->psi_;
      alpha_     = in->alpha_;
      knSq_      = in->knSq_;
      ksSq_      = in->ksSq_;
      etanSq_    = in->etanSq_;
      etasSq_    = in->etasSq_;
      moisState_ = in->moisState_;
      del0_      = in->del0_;
      Fs_        = in->Fs_;
      Fm_        = in->Fm_;
      //
      Fsh0_ = in->Fsh0_;
      R0_   = in->R0_;
    }

    bool ContactModelHill::forceDisplacementLaw(ContactModelMechanicalState *state,const double &timestep) {

      // (1) Update the moisture force.
      if ( moisState_ != 2 )
        Fm_ = 0.0;
      else {
        double Two_scr = 2.0 * R0_;
        double gc = state->gap_;
        Fm_ = 0.0;
        if ( gc <= Two_scr ) {
          double Fm_max = psi_ * M_PI * R0_ * R0_;
          if ( gc < 0.0 ) Fm_ = Fm_max;
          else            Fm_ = Fm_max * exp( -gc / Two_scr );
        }
        if (    (moisState_ == 2)
             && ( gc > Two_scr )  ) {
          moisState_ = 1;  // rupture the liquid bridge
        }
      }

      // (2) Update the normal surface-interaction force.
      del0_ = -state->gap_;
      double Fns( 0.0 );
      if ( del0_ <= 0.0 )
        Fns = 0.0;
      else {
        double del0_dot = -state->relativeTranslationalIncrement_.rdof(0) / timestep;
        Fns =    -knSq_ * pow(del0_, 1.50)
              - etanSq_ * pow(del0_, 0.25) * del0_dot;
      }
      Fs_.rdof(0) = -Fns;

      // (3) Update the shear force.
      DVect2 Fsh( 0.0 ); // Hertzian shear force
      DVect2 Fs(  0.0 ); // shear force
      if ( del0_ > 0.0 ) {
        double Fs_mu = -mu_ * Fns;
        DVect2 Dds( state->relativeTranslationalIncrement_.rdof(1),
                    state->relativeTranslationalIncrement_.rdof(2)  );
        DVect2 Fsh_star =   Fsh0_
                          - Dds * ksSq_ * pow(del0_, 0.5);
        if ( Fsh_star.mag() <= Fs_mu ) Fsh = Fsh_star;
        else                           Fsh = Fsh_star.unit() * Fs_mu;
        //
        if ( etasSq_ == 0.0 ) Fs = Fsh;
        else { // add dashpot contribution
          DVect2 ds_dot( state->relativeTranslationalIncrement_.rdof(1) / timestep,
                         state->relativeTranslationalIncrement_.rdof(2) / timestep  );
          DVect2 Fs_star =   Fsh
                           - ds_dot * etasSq_ * pow(del0_, 0.25);
          if ( Fs_star.mag() <= Fs_mu ) Fs = Fs_star;
          else                          Fs = Fs_star.unit() * Fs_mu;
        }
      }
      Fsh0_ = Fsh;  // store value for use during next time step
      Fs_.rdof(1) = Fs.rdof(0);
      Fs_.rdof(2) = Fs.rdof(1);

      // Update the contact force and moment.
      state->force_.rdof(0) = Fs_.rdof(0) - Fm_;
      state->force_.rdof(1) = Fs_.rdof(1);
      state->force_.rdof(2) = Fs_.rdof(2);
      state->momentOn1_.fill(0.0);
      state->momentOn2_.fill(0.0);
      state->getMechanicalContact()->updateResultingTorquesLocal( state->force_,
                                                                  &state->momentOn1_, &state->momentOn2_ );
      return checkActivity( state->gap_ );
    }

    DVect2 ContactModelHill::getEffectiveTranslationalStiffness() const {
      double del0 = ( del0_ <= 0.0 ? 2.0e-4*R0_ : del0_ );  // 0.01% of minimum diameter of two contacting pieces
      double kn = 1.5 * knSq_ * pow(del0, 0.5);
      double ks =       ksSq_ * pow(del0, 0.5);
      //
      if (alpha_ > 0.0) {  // Account for active dashpots.
        double Beta = 0.5 * alpha_;  // assumption, most likely incorrect (should be okay when alpha_ <= 0.1)
        double F = sqrt( 1.0 + Beta*Beta ) - Beta;
        kn += kn / (F*F) ;
        ks += ks / (F*F) ;
      }
      return( DVect2( kn, ks ) );
    }

    bool ContactModelHill::validate( ContactModelMechanicalState *state, const double & ) {
      assert(state);
      const IContactMechanical *c = state->getMechanicalContact();
      assert(c);
      updateCoefs(c);
      return checkActivity( state->gap_ );
    }

    bool ContactModelHill::checkActivity( const double &gap ) {
      return checkActivityConst( gap );
    }

    bool ContactModelHill::checkActivityConst( const double &gap ) const {
      if ( moisState_ == 2 ) return true;
      return ( gap <= 0 ? true : false );
    }

    bool ContactModelHill::isOKToDelete() const {
      // Assume that gap equals -del0_ (may no longer be the case, but only off by one step).
      // We allow inactive contacts to be deleted. We should consider inhibiting deletion if the gap
      // is less than the moisture gap --- but for now, the moisture gap is not a property of the hill model.
      return !checkActivityConst( -del0_ );
    }

    QString  ContactModelHill::getMethodArguments(uint i) const {
      switch (i) {
        case kwMakeWet:
          return "gap";
      }
      assert(0);
      return QString();
    }

    bool ContactModelHill::setMethod( uint i, const QVector<QVariant> &vl, IContact *con ) {
      IContactMechanical *c( convert_getcast<IContactMechanical>(con) );
      switch (i) {
        case kwMakeWet: {
          double gm;
          if ( vl.at(0).isNull() ) 
            throw Exception( "Argument gap must be specified with make_wet method of"
                             " the %1 contact model.", getName() );
          gm = vl.at(0).toDouble();
          if ( gm < 0.0 )
            throw Exception( "Negative moisture gap not allowed with make_wet method of"
                             " the %1 contact model.", getName() );
          if ( c->getGap() <= gm )
            moisState_ = 2;
          break;
        }
        case kwMakeDry: {
          moisState_ = 0;
          break;
        }
      }
      return false;
    }

    static const QString YoungModStr( "young_mod" );
    static const QString PoisRatioStr("pois_ratio");
    void ContactModelHill::updateCoefs( const IContactMechanical *con ) {
      assert(con);
      QVariant v1, v2;
      // ********** Get {E1, E2}.
      v1 = con->getEnd1()->getProperty(YoungModStr);
      v2 = con->getEnd2()->getProperty(YoungModStr);
      if (!v1.isValid() || !v2.isValid())
        throw Exception( "Surface property young_mod has not been assigned to a body that interacts with"
                         " the %1 contact model.", getName() );
      double E1 = v1.toDouble();
      double E2 = v2.toDouble();
      if ( (E1 <= 0.0) || (E2 <= 0.0) )
          throw Exception( "Surface property young_mod is not positive for a body that interacts with"
                           " the %1 contact model.", getName() );
      // ********** Get {nu1, nu2}.
      v1 = con->getEnd1()->getProperty(PoisRatioStr);
      v2 = con->getEnd2()->getProperty(PoisRatioStr);
      if (!v1.isValid() || !v2.isValid())
        throw Exception( "Surface property pois_ratio has not been assigned to a body that interacts with"
                         " the %1 contact model.", getName() );
      double nu1 = v1.toDouble();
      double nu2 = v2.toDouble();
      if (    (nu1 <= -1.0) || (nu1 > 0.5)
           || (nu2 <= -1.0) || (nu2 > 0.5) )
          throw Exception( "Surface property pois_ratio is not in range (-1.0, 0.5] for a body that interacts with"
                           " the %1 contact model.", getName() );
      // ********** Get {R0_, Re, Ee, Ge, me}.
      if ( !con->getEnd1Curvature().rdof(1) || !con->getEnd2Curvature().rdof(1) )
          throw Exception( "Cannot get radius of a body that interacts with"
                           " the %1 contact model.", getName() );
      double R1 = 1.0 / con->getEnd1Curvature().rdof(1);      
      double R2 = 1.0 / con->getEnd2Curvature().rdof(1);      
      R0_ = min( R1, R2 );
      double Re = 1.0 / ( con->getEnd1Curvature().rdof(1) + con->getEnd2Curvature().rdof(1) );
      double Ee = 1.0 / (  (1.0 - nu1*nu1)/E1
                         + (1.0 - nu2*nu2)/E2 );
      double Ge = 1.0 / (   ( 2.0*(1.0 + nu1)*(2.0 - nu1) )/E1
                          + ( 2.0*(1.0 + nu2)*(2.0 - nu2) )/E2 );
      const IBody *b1 = con->getEnd1()->getIBody();
      const IBody *b2 = con->getEnd2()->getIBody();
      double m1 = convert_getcast<IBodyMechanical>(b1)->getInertialMass();
      double m2 = convert_getcast<IBodyMechanical>(b2)->getInertialMass();
      if ( (m1 <= 0.0) || (m2 <= 0.0) )
          throw Exception( "The mass is not positive for a body that interacts with"
                           " the %1 contact model.", getName() );
      double me = m1 * m2 / (m1 + m2);
      // ********** Now, update the stiffness and damping coefficients.
      knSq_ = (4.0 / 3.0) * Ee * sqrt(Re);
      ksSq_ =         8.0 * Ge * sqrt(Re);
      etanSq_ = alpha_ * sqrt( me * knSq_ );
      etasSq_ = alpha_ * sqrt( me * ksSq_ );
    }

} // namespace cmodelsxd
// EoF