#pragma once
// contactmodelHill.h

#include "contactmodel/src/contactmodelmechanical.h"

#ifdef Hill_LIB
#  define Hill_EXPORT EXPORT_TAG
#elif defined(NO_MODEL_IMPORT)
#  define Hill_EXPORT
#else
#  define Hill_EXPORT IMPORT_TAG
#endif

namespace cmodelsxd {
    using namespace itasca;

    class ContactModelHill : public ContactModelMechanical {
      public:
        Hill_EXPORT               ContactModelHill();
                                  // Constructor: Set default values for contact model properties.
        Hill_EXPORT virtual      ~ContactModelHill();
                                  // Destructor, called when contact is deleted: free allocated memory, etc.
        virtual ContactModelHill *clone() const { return NEWC(ContactModelHill()); }
  
        virtual QString  getName() const  { return "hill"; }
                         // Contact model name (used as keyword for commands and FISH).
        virtual uint     getMinorVersion() const;            
                         // Contact model version number. The version number can be
                         // accessed during the save-restore operation (within the archive method,
                         // testing {stream.getRestoreVersion() == getMinorVersion()}) to allow for 
                         // future modifications to the contact model data structure. The version
                         // number is given by the command {contact list modelllist}.
        virtual void     setIndex(int i)  { index_=i; }
        virtual int      getIndex() const { return index_; }
                         // The index provides a quick way to determine the type of contact model.
                         // Each type of contact model in PFC must have a unique index; this is assigned
                         // by PFC when the contact model is loaded. This index should be set to -1.
      
        virtual QString  getProperties() const { // Contact model property names. 
          return  "fric_coef"
                 ",suction"
                 ",damp_con"
                 ",young_mod"
                 ",pois_ratio"
                 ",nstiff_coef"
                 ",sstiff_coef"
                 ",ndamp_coef"
                 ",sdamp_coef"
                 ",overlap"
                 ",mois_state"
                 ",surf_force"
                 ",mois_force";
        }
        enum PropertyKeys { 
                   kwFricCoef=1                            
                 , kwSuction
                 , kwDampCon   
                 , kwYoungMod
                 , kwPoisRatio
                 , kwNStiffCoef
                 , kwSStiffCoef
                 , kwNDampCoef
                 , kwSDampCoef
                 , kwOverlap
                 , kwMoisState
                 , kwSurfForce
                 , kwMoisForce
        };
        virtual bool     getPropertyReadOnly(uint i) const;
                         // The return value denotes whether or not the property is read-only
                         // (TRUE: read-only, FALSE: read-write).
        virtual bool     getPropertyGlobal(uint i) const;
                         // The return value denotes whether or not the property corresponds to the global
                         // or local coordinate system (TRUE: global system, FALSE: local system). The
                         // local system is the contact-plane system (nst) defined as follows.
                         // If a vector V is expressed in the local system as (Vn, Vs, Vt), then V is
                         // expressed in the global system as {Vn*nc + Vs*sc + Vt*tc} where where nc, sc
                         // and tc are unit vectors in directions of the nst axes.
                         // This is used when rendering contact model properties that are vectors.
        virtual bool     supportsInheritance(uint i) const;
                         // The return value denotes whether or not the property is inheritable
                         // (TRUE: inheritable, FALSE: not inheritable). Inheritance is provided by
                         // the endPropertyUpdated method.
        virtual bool     getInheritance(uint i) const { return true; }
        virtual void     setInheritance(uint i,bool b) { }
        virtual bool     endPropertyUpdated( const QString &name, const IContactMechanical *c );
                         // The endPropertyUpdated function is called whenever a surface property (with a name
                         // that matches an inheritable contact model property name) of one of the contacting
                         // pieces is modified. This allows the contact model to update its associated
                         // properties. The return value denotes whether or not the update has affected
                         // the time step computation (by having modified the translational or rotational
                         // tangent stiffnesses). If true is returned, then the time step will be recomputed.  
        virtual QVariant getProperty(uint i,const IContact *) const;
                         // Return the specified contact model property.
        virtual bool     setProperty(uint i,const QVariant &v,IContact *);
                         // Set the specified contact model property, ensuring that it is of the correct type
                         // and within the correct range --- if not, then throw an exception.
                         // The return value denotes whether or not the update has affected the time step
                         // computation (by having modified the translational or rotational tangent stiffnesses).
                         // If true is returned, then the time step will be recomputed.
        virtual void     archive(ArchiveStream &);
                         // Provide save-restore capability for the state information.
        virtual void     copy(const ContactModel *c);
                         // Copy the state information to a newly created contact model.
                         // Provide access to state information, for use by copy method.
        virtual bool     forceDisplacementLaw( ContactModelMechanicalState *state, const double &timestep );
                         // The forceDisplacementLaw function is called during each cycle. Given the relative
                         // motion of the two contacting pieces (via
                         //   state->relativeTranslationalIncrement_ (Ddn, Ddss, Ddst)
                         //   state->relativeAngularIncrement_       (Dtt, Dtbs, Dtbt)
                         //     Ddn  : relative normal-displacement increment, Ddn > 0 is opening
                         //     Ddss : relative  shear-displacement increment (s-axis component)
                         //     Ddst : relative  shear-displacement increment (t-axis component)
                         //     Dtt  : relative twist-rotation increment
                         //     Dtbs : relative  bend-rotation increment (s-axis component)
                         //     Dtbt : relative  bend-rotation increment (t-axis component)
                         //       The relative displacement and rotation increments:
                         //         Dd = Ddn*nc + Ddss*sc + Ddst*tc
                         //         Dt = Dtt*nc + Dtbs*sc + Dtbt*tc
                         //       where nc, sc and tc are unit vectors in direc. of the nst axes, respectively.
                         //       [see {Table 1: Contact State Variables} in PFC Model Components:
                         //       Contacts and Contact Models: Contact Resolution]
                         // ) and the contact properties, this function must update the contact force and
                         // moment (via state->force_, state->momentOn1_, state->momentOn2_).
                         //   The force_ is acting on piece 2, and is expressed in the local coordinate system
                         //   (defined in getPropertyGlobal) such that the first component positive denotes
                         //   compression. If we define the moment acting on piece 2 by Mc, and Mc is expressed
                         //   in the local coordinate system (defined in getPropertyGlobal), then we must set
                         //   {state->momentOn1_ = -Mc}, {state->momentOn2_ = Mc} and call
                         //   state->getMechanicalContact()->updateResultingTorquesLocal(...) after having set
                         //   force_.
                         // The return value indicates the contact activity status (TRUE: active, FALSE:
                         // inactive) during the next cycle.
                         // Additional information:
                         //   * If state->activated() is true, then the contact has just become active (it was
                         //     inactive during the previous time step).
                         //   * Fully elastic behavior is enforced during the SOLVE ELASTIC command by having
                         //     the forceDispLaw handle the case of {state->canFail_ == true}.
 
        virtual DVect2   getEffectiveTranslationalStiffness() const;
        virtual DAVect   getEffectiveRotationalStiffness() const    { return DAVect(0.0);      }
                         // The getEffectiveXStiffness functions return the translational and rotational
                         // tangent stiffnesses used to compute a stable time step. When a contact is sliding,
                         // the translational tangent shear stiffness is zero (but this stiffness reduction
                         // is typically ignored when computing a stable time step). If the contact model
                         // includes a dashpot, then the translational stiffnesses must be increased (see
                         // Potyondy (2009)).
                         //   [Potyondy, D. Stiffness Matrix at a Contact Between Two Clumps,  Itasca
                         //   Consulting Group, Inc., Minneapolis, MN, Technical Memorandum ICG6863-L,
                         //   December 7, 2009.]
        virtual bool     validate( ContactModelMechanicalState *state, const double &timestep );
                         // Prepare for entry into ForceDispLaw. The validate function is called when:
                         // (1) the contact is created, (2) a property of the contact that returns a true via
                         // the setProperty method has been modified and (3) when a set of cycles is executed
                         // via the {cycle N} command.
                         // Return value indicates contact activity (TRUE: active, FALSE: inactive).
        virtual bool     checkActivity( const double &gap );
                         // The checkActivity function is called by the contact-resolution logic when...
                         // Return value indicates contact activity (TRUE: active, FALSE: inactive).
                         // A contact with the hill model is active if it is wet, or if the contact gap is
                         // less than or equal to zero.
        virtual double   getActivityDistance() const { return 0.0; }
                         // The getActivityDistance function is called by the contact-resolution logic when
                         // the CMAT is modified. Return value is the activity distance used by the
                         // checkActivity function. For the hill model, this is zero, because we
                         // provide a moisture gap via the proximity of the CMAT followed by CLEAN command.
        virtual bool     isOKToDelete() const;
                         // The isOKToDelete function is called by the contact-resolution logic when...
                         // Return value indicates whether or not the contact may be deleted.
                         // If TRUE, then the contact may be deleted when it is inactive.
                         // If FALSE, then the contact may not be deleted (under any condition).
        virtual void     resetForcesAndMoments() { Fs_ = DVect3( 0.0 ); Fm_ = 0.0; }
                         // Zero the forces and moments stored in the contact model. This function is called
                         // when the contact becomes inactive.
        
        enum MethodKeys  { kwMakeWet=1, kwMakeDry };
        virtual QString  getMethods() const { return "make_wet,make_dry"; }
        virtual QString  getMethodArguments(uint i) const; 
        virtual bool     setMethod( uint i, const QVector<QVariant> &vl, IContact *con=0 );
                         // The return value denotes whether or not the operation has affected the time step
                         // computation (by having modified the translational or rotational tangent stiffnesses).
                         // If true is returned, then the time step will be recomputed.

    private:
        static int index_;

        // hill model properties
        double mu_;        // friction coef.
        double psi_;       // suction [stress]
        double alpha_;     // damping constant
        double knSq_;      // normal stiffness coef. [N * m^{-3/2}]
        double ksSq_;      // shear  stiffness coef. [N * m^{-3/2}]
        double etanSq_;    // normal damping coef. [kg * m^{-1/4} * s^{-1}]
        double etasSq_;    // shear  damping coef. [kg * m^{-1/4} * s^{-1}]
        int    moisState_; // moisture state (0: dry, 1: dry & ruptured, 2: wet)
        double del0_;      // contact overlap (del0_ > 0 is overlap)
        DVect3 Fs_;        // surface-interaction force (local coord. system, first component positive denotes compression)
        double Fm_;        // moisture force (Fm_ > 0 is suction)

        // Private state variables.
        DVect2 Fsh0_; // Hertzian shear force (local coord. system) at beginning of time step
        double R0_;   // minimum radius of two contacting pieces

        void updateCoefs( const IContactMechanical *con );
             // Update the stiffness and damping coefficients, and R0_.
        bool checkActivityConst( const double &gap ) const;

    };
} // namespace cmodelsxd
// EoF

